from sys import argv

from backscattering_analyzer.settings import Settings
from science_signal.signal import Signal
from science_signal import interpolate
from backscattering_analyzer import (
    compute_light,
    fit_compute_light,
    opt_compute_light,
)
from numpy import argmin, loadtxt, logspace, pi, array
from scipy.io.matlab import loadmat


class Analyzer:
    """
    Utility class to study backscattering light in VIRGO
    """

    def __init__(
        self, arguments: None | list[str] | str = None
    ) -> None:
        if arguments is None:
            options = []
        elif isinstance(arguments, str):
            options = arguments.split(" ")
        else:
            options = arguments
        self.settings = Settings(argv[1:] + options)
        self.load()
        self.process_movement()
        self.compute_light()

    def load(self):
        """
        Load all data
        """
        self.settings.log("Load all data")
        try:
            self.load_bench()
            self.load_mirror()
            self.load_data()
            self.load_reference()
            self.load_coupling()
        except Exception:
            raise Exception("An error occured during data loading")

    def load_bench(self):
        """
        Load bench movement
        """
        file = self.settings.bench_file()
        self.settings.log("loading bench movement: {}".format(self.settings.bench_file()))
        try:
            data = loadtxt(file).T
            self.bench_movement = Signal(data[0], data[1]) * 1e-6  # um
        except OSError:
            raise Exception("{file} does not exist".format(file=file))

    def load_mirror(self):
        """
        Load mirror movement
        """
        file = self.settings.mirror_file()
        self.settings.log("loading mirror movement: {}".format(self.settings.mirror_file()))
        try:
            data = loadtxt(file).T
            self.mirror_movement = Signal(data[0], data[1]) * 1e-6 # um
        except OSError:
            raise Exception("{file} does not exist".format(file=file))

    def load_data(self):
        """
        Load excited h(t)
        """
        file = self.settings.data_file()
        self.settings.log("loading excited h(t): {}".format(self.settings.data_file()))
        try:
            data = loadtxt(file).T
            self.data_signal = Signal(data[0], data[1])
        except OSError:
            raise Exception("{file} does not exist".format(file=file))

    def load_reference(self):
        """
        Load reference h(t)
        """
        file = self.settings.reference_file()
        self.settings.log("loading reference h(t): {}".format(self.settings.reference_file()))
        try:
            data = loadtxt(file).T
            self.reference_signal = Signal(data[0], data[1])
        except OSError:
            raise Exception("{file} does not exist".format(file=file))

    def load_coupling(self):
        """
        Load modelisation coupling data
        """
        file = self.settings.modelisation_file()
        self.settings.log("loading matlab modelisation data: {}".format(self.settings.modelisation_file()))
        try:
            self.modelisation = loadmat(file)
            coupling_values = self.modelisation[
                self.settings.coupling_name()
            ]
            self.coupling = [
                Signal(
                    self.modelisation["freq"][0],
                    coupling,
                )
                for coupling in coupling_values
            ]
        except OSError:
            raise Exception("{file} does not exist".format(file=file))

    def process_movement(self):
        """
        Clean and compute relative movement between the bench and the
        miror
        """
        self.settings.log("computing relative movement")

        self.movement = (
            (
                self.bench_movement * self.settings.calib_bench
                - self.mirror_movement * self.settings.calib_mirror
            )
            .detrend("linear")
            .filter(end=5 * self.settings.vibration_frequency)
        )
    
    def compute_light(self) -> None:
        """
        Compute projection with current bench excitation
        """
        self.settings.log("computing backscatterd light projection")
        self.projection = compute_light(
            scatter_factor=self.settings.scattering_factor,
            coupling=self.coupling,
            movement=self.movement,
            wavelength=self.settings.wavelength,
            power_in=self.settings.power_in,
            power_out=self.settings.power_out,
        )

    def fit_scatter_factor(self, guess: None | float = None) -> float:
        """
        Find the best scatter factor (first order only) in the given
        range
        """
        if guess is None:
            guess = self.settings.scattering_factor[0]

        import matplotlib.pyplot as plt

        phase = 4 * pi / self.settings.wavelength
        factor_n = (self.movement * phase).sin().psd().sqrt()
        coupling_n = abs(self.coupling[0])
        factor_d = (self.movement * phase).cos().psd().sqrt()
        coupling_d = abs(self.coupling[1])

        coupling_d = coupling_d.cut(
            10,
            200,
        )  # cut signal between 10 and 200 Hz (updated to 11-15 to test)

        factor_n, coupling_n, factor_d, coupling_d, data, reference = (
            interpolate(
                factor_n,
                coupling_n,
                factor_d,
                coupling_d,
                self.data_signal.psd().sqrt(),
                self.reference_signal.psd().sqrt(),
            )
        )

        x = logspace(-15, -3, 1000)
        y = array(
            [
                fit_compute_light(
                    x[i],
                    factor_n,
                    coupling_n,
                    factor_d,
                    coupling_d,
                    self.settings.power_in,
                    self.settings.power_out,
                    data,
                    reference,
                    phase,
                )
                for i in range(len(x))
            ]
        )

        factor: float = x[argmin(y)]

        projection = Signal(
            factor_n.x,
            opt_compute_light(
                scatter_factor=factor,
                factor_n=factor_n,
                coupling_n=coupling_n,
                factor_d=factor_d,
                coupling_d=coupling_d,
                power_in=self.settings.power_in,
                power_out=self.settings.power_out,
                phase=phase,
            ),
        )
        _ = plt.loglog(x, y)
        _ = plt.show()

        _ = plt.loglog(projection.x, projection.y, label="projection")
        _ = plt.loglog(reference.x, reference.y, label="référence")
        _ = plt.loglog(data.x, data.y, label="data")
        _ = plt.loglog(
            reference.x, projection.y + reference.y, label="somme"
        )
        _ = plt.legend()
        _ = plt.show()

        return factor
